# ===============================================
# (C) 2018 AO Kaspersky Lab. All Rights Reserved.
# ===============================================
import binascii
import hashlib
import logging

import dt_load_package_libs
assert dt_load_package_libs

import pandas as pd
import sqlalchemy

from dt_checks import is_bad_domain, is_unpopular_domain_ksn
import dt_utils

from dt_Options import AlertHasherOptions as options


@dt_utils.reference_data('GetNewExtAlerts', pd.Timestamp.min, pd.Timestamp.max)

def get_new_ext_alerts(connection, date_from, date_till):
    logging.info('Processing interval: from=%s, till=%s', date_from, date_till)
    logging.debug('Loading new ext alerts from DB...')
    return dt_utils.get_multiple_data(connection, get_new_ext_alerts.proc, date_from, date_till)


def get_hashes_domain(new_ext_alerts):
    df = new_ext_alerts['textalerturl']

    df = df[is_bad_domain(df, options.HASH_HOST_CATEGORIES) |
            is_unpopular_domain_ksn(df)]

    df = df[['extalertid', 'domainshortname']].copy()

    df['alerthashtypename'] = 'domain'
    df['alerthash'] = [hashlib.md5(d.lower().encode('utf-8')).digest()
                       for d in df['domainshortname']]

    return df.rename(columns={'domainshortname': 'alerthashvalue'})


def get_hashes_filemd5(new_ext_alerts):
    df = new_ext_alerts['textalertfile']

    df = df[ df['lastupdatetime'].notnull() &
            (df['wocuserscount'].fillna(0) < 500) &
            (df['mixuserscount'].fillna(0) < 500) &
             df['hipsfilegroup'].isin([3, 255]) &
            (df['vhsgoodfilescount'].fillna(0) == 0)]

    df = df[['extalertid', 'filemd5']].copy()
    df['alerthashtypename'] = 'filemd5'
    df['alerthashvalue'] = [binascii.hexlify(f) for f in df['filemd5']]
    return df.rename(columns={'filemd5': 'alerthash'})


def get_hashes_filename(new_ext_alerts):
    df = new_ext_alerts['textalertfile']

    df = df[ df['filelastupdatetime'].notnull() &
            (df['filenameuserscount'].fillna(0) < 100)]

    df = df[['extalertid', 'filename']].copy()
    df['alerthashtypename'] = 'filename'
    df['alerthash'] = [hashlib.md5(fn.lower().encode('utf-8')).digest()
                       for fn in df['filename']]
    return df.rename(columns={'filename': 'alerthashvalue'})


def get_hashes_filepath(new_ext_alerts):
    df = new_ext_alerts['textalertfile']

    df = df[ df['lastupdatetime_filepathnorm'].notnull() &
            (df['ksnusers_filepathnorm'].fillna(0) < 100)]

    df = df[['extalertid', 'filepathnorm']].copy()
    df['alerthashtypename'] = 'filepath'
    df['alerthash'] = [hashlib.md5(fp.lower().encode('utf-8')).digest()
                       for fp in df['filepathnorm']]
    return df.rename(columns={'filepathnorm': 'alerthashvalue'})


def get_hashes_url(new_ext_alerts):
    df = new_ext_alerts['textalerturl']
    df = df[df['url'].notnull()]

    df = df[is_bad_domain(df, options.HASH_HOST_CATEGORIES) |
            is_unpopular_domain_ksn(df)]

    df = df[['extalertid', 'url']].copy()

    df['alerthashtypename'] = 'url'
    df['alerthash'] = [hashlib.md5(u.encode('utf-8')).digest() for u in df['url']]
    return df.rename(columns={'url': 'alerthashvalue'})


def get_hashes_vhs(new_ext_alerts):
    df = new_ext_alerts['textalertfile']

    df = df[ df['filevhs'].notnull() &
             df['vhslastupdatetime'].notnull() &
            (df['vhsnetcompcount'].fillna(0) < 10) &
            (df['vhsgoodfilescount'].fillna(0) == 0) &
            ((df['vhsuserscount'].fillna(0) < 500) | (df['vhsbadfilescount'] > 0))]

    df = df[['extalertid', 'filevhs']].copy()
    df['alerthashtypename'] = 'vhs'
    df['alerthashvalue'] = [binascii.hexlify(v) for v in df['filevhs']]
    return df.rename(columns={'filevhs': 'alerthash'})


def update_alert_hashes(connection, alert_hashes):
    logging.debug('Saving alert hashes to DB')

    if dt_utils.is_postgresql(connection):
        schema, table_name = 'pg_temp', 'talerthashes'
    else:
        schema, table_name = None, '#alerthashes'

    alert_hashes.drop_duplicates(['alerthash', 'alerthashtypename', 'extalertid'], inplace=True)

    dtype = {'alerthash': sqlalchemy.dialects.postgresql.BYTEA}
    alert_hashes.to_sql(table_name, con=connection, index=False, schema=schema, dtype=dtype)
    connection.execute(sqlalchemy.text('SELECT alert.AddAlertHashes()'))


def process_data(connection, new_ext_alerts, date_from, date_till):
    n_ext_alerts = sum(len(df) for df in new_ext_alerts.values())
    logging.info('Got %d new ext alerts', n_ext_alerts)

    alert_hashes = pd.concat([get_hashes_domain(new_ext_alerts),
                              get_hashes_filemd5(new_ext_alerts),
                              get_hashes_filepath(new_ext_alerts),
                              get_hashes_url(new_ext_alerts),
                              get_hashes_vhs(new_ext_alerts)])

    alert_hashes['alertid'] = float('nan')
    logging.info('Created %d alerts hashes', len(alert_hashes))

    if len(alert_hashes) > 0:
        hash_type_stats = []
        for hash_type_name, count in alert_hashes.groupby('alerthashtypename').size().iteritems():
            hash_type_stats.append('#{0}={1}'.format(hash_type_name, count))

        logging.debug('New alert hashes: %s', ', '.join(hash_type_stats))

    with connection.begin():
        if len(alert_hashes) > 0:
            update_alert_hashes(connection, alert_hashes)

        logging.debug('Updating from/till interval')
        dt_utils.set_processed_date(connection, options.DB_KEY, date_till)

    logging.info('Processing finished.')
    return True


if __name__ == '__main__':
    dt_utils.test_module(dt_utils.get_test_module_args(), options, get_new_ext_alerts)
    exit(dt_utils.run_module(options, get_new_ext_alerts, process_data))
